import os

def get_linked_node_and_socket(node_tree, socket):
	for l in node_tree.links:
		if socket==l.to_socket:
			return (l.from_node, l.from_socket)
		elif socket==l.from_socket:
			return (l.to_node, l.to_socket)
	return (None, None)

class MaterialProperty:
	def __init__(self, value):
		self.value = value
		self.texture = None
		self.tex_usage = None

	def set_from_input(self, node_tree, input_socket, alpha_socket=None):
		if type(self.value)==tuple:
			if alpha_socket:
				self.value = input_socket.default_value[:len(self.value)-1]+(alpha_socket.default_value,)
			else:
				self.value = input_socket.default_value[:len(self.value)]
		elif self.value is not None:
			self.value = input_socket.default_value

		from_node, _ = get_linked_node_and_socket(node_tree, input_socket)
		alpha_from = None
		if from_node:
			if from_node.type=='NORMAL_MAP':
				from_node, _ = get_linked_node_and_socket(node_tree, from_node.inputs["Color"])

			if alpha_socket:
				alpha_from, _ = get_linked_node_and_socket(node_tree, alpha_socket)
				if alpha_from and alpha_from!=from_node:
					raise Exception("Separate textures for color and alpha are not supported")

			if from_node.type=='TEX_IMAGE':
				self.texture = from_node
				if alpha_from:
					self.tex_usage = 'RGBA'
				elif type(self.value)==tuple:
					self.tex_usage = 'RGB'
				else:
					self.tex_usage = 'GRAY'
			else:
				raise Exception("Unsupported property input node type "+from_node.type)

class Material:
	def __init__(self, material):
		self.name = material.name
		self.type = None
		self.properties = {}

		self.render_mode = material.render_mode
		self.technique = material.technique
		self.shader = material.shader

		if self.render_mode=='EXTERNAL' and not self.technique:
			raise Exception("Missing technique with external rendering mode")
		elif self.render_mode=='CUSTOM' and not self.shader:
			raise Exception("Missing shader with custom rendering mode")

		out_node = None
		for n in material.node_tree.nodes:
			if n.type=='OUTPUT_MATERIAL':
				out_node = n
				break

		if not out_node:
			raise Exception("No material output node found")

		surface_node, _ = get_linked_node_and_socket(material.node_tree, out_node.inputs["Surface"])
		if not surface_node:
			if self.render_mode=='BUILTIN':
				raise Exception("Empty material can't use builtin rendering mode")
			return
		elif surface_node.type!='BSDF_PRINCIPLED':
			raise Exception("Unsupported surface node type "+surface_node.type)

		self.type = "pbr"

		base_color = self.properties["base_color"] = MaterialProperty((0.8, 0.8, 0.8, 1.0))
		metalness = self.properties["metalness"] = MaterialProperty(0.0)
		roughness = self.properties["roughness"] = MaterialProperty(0.5)
		normal = self.properties["normal"] = MaterialProperty(None)
		emission = self.properties["emission"] = MaterialProperty((0.0, 0.0, 0.0))

		base_color.set_from_input(material.node_tree, surface_node.inputs["Base Color"], surface_node.inputs["Alpha"])
		metalness.set_from_input(material.node_tree, surface_node.inputs["Metallic"])
		roughness.set_from_input(material.node_tree, surface_node.inputs["Roughness"])
		normal.set_from_input(material.node_tree, surface_node.inputs["Normal"])
		emission.set_from_input(material.node_tree, surface_node.inputs["Emission"])

		sampler_settings = None
		for p in self.properties.values():
			if p.texture:
				settings = (p.texture.default_filter, p.texture.interpolation, p.texture.use_mipmap, p.texture.max_anisotropy)
				if sampler_settings is None:
					sampler_settings = settings
				elif settings!=sampler_settings:
					raise Exception("Conflicting sampler settings in material textures")


class MaterialMap:
	def __init__(self, materials):
		self.render_mode = materials[0].render_mode
		if self.render_mode=='EXTERNAL':
			raise Exception("Material map with external render mode does not make sense")

		self.shader = materials[0].shader
		if self.shader:
			self.name = "material_map_"+os.path.splitext(self.shader)[0]
		else:
			self.name = "material_map"
		self.materials = materials
		self.material_names = [m.name for m in self.materials]
		for m in self.materials:
			if m.render_mode!=self.render_mode:
				raise Exception("Conflicting render modes in MaterialMap constructor")
			if self.render_mode=='CUSTOM' and m.shader!=self.shader:
				raise Exception("Conflicting shaders in MaterialMap constructor")

		count = len(self.materials)
		size = 1
		while size*size*2<count:
			size *= 2
		if size*size>=count:
			self.size = (size, size)
		else:
			self.size = (size*2, size)

		from .util import get_colormap

		cm = get_colormap(True)
		self.base_color_data = ""
		for m in map(Material, self.materials):
			if any(p.texture for p in m.properties):
				raise Exception("Texturing is incompatible with material map")
			base_color = [int(cm(c)*255) for c in m.base_color.value]
			self.base_color_data += "\\x{:02X}\\x{:02X}\\x{:02X}\\xFF".format(*base_color)
		self.base_color_data += "\\x00\\x00\\x00\\x00"*(self.size[0]*self.size[1]-count)

	def get_material_uv(self, material):
		index = self.material_names.index(material.name)
		x = index%self.size[0]
		y = index//self.size[0]
		return ((x+0.5)/self.size[0], (y+0.5)/self.size[1])

def create_material_map(context, material):
	if not material.material_map:
		raise Exception("Material is not part of a material map")

	shader = material.shader
	materials = []
	for m in context.blend_data.materials:
		if m.material_map and m.shader==shader:
			materials.append(m)

	mat_map = MaterialMap(materials)

	return mat_map
