#include <msp/stringcodec/utf8.h>
#include <msp/strings/format.h>
#include "debug.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

const std::string &DumpTree::apply(Stage &stage)
{
	formatted = format("Stage: %s\n", Stage::get_stage_name(stage.type));
	tree.push_back(BRANCH);
	append(format("Version: %d.%02d", stage.required_features.glsl_version.major, stage.required_features.glsl_version.minor));

	for(std::map<string, TypeDeclaration *>::const_iterator i=stage.types.begin(); i!=stage.types.end(); ++i)
		append(format("Type: %%%d %s", get_label(*i->second), i->first));

	set<InterfaceBlock *> seen_interfaces;
	for(std::map<string, InterfaceBlock *>::const_iterator i=stage.interface_blocks.begin(); i!=stage.interface_blocks.end(); ++i)
		if(seen_interfaces.insert(i->second).second)
		{
			string text = format("Interface block: %%%d %s %s", get_label(*i->second), i->second->interface, i->second->name);
			if(!i->second->instance_name.empty())
				text += format(" %s", i->second->instance_name);
			append(text);
		}

	for(std::map<string, FunctionDeclaration *>::const_iterator i=stage.functions.begin(); i!=stage.functions.end(); ++i)
		append(format("Function: %%%d %s", get_label(*i->second), i->first));

	last_branch();
	stage.content.visit(*this);
	return formatted;
}

void DumpTree::append(const string &line)
{
	StringCodec::Utf8::Encoder enc;
	for(vector<TreeChars>::const_iterator i=tree.begin(); i!=tree.end(); )
	{
		enc.encode_char(*i++, formatted);
		enc.encode_char((i==tree.end() ? REACH : EMPTY), formatted);
	}
	formatted += line;
	formatted += '\n';
}

void DumpTree::begin_sub()
{
	tree.back() = (tree.back()==BRANCH_LAST ? EMPTY : STRAIGHT);
	tree.push_back(BRANCH);
}

void DumpTree::last_branch()
{
	tree.back() = BRANCH_LAST;
}

void DumpTree::end_sub()
{
	tree.pop_back();
	if(tree.back()==STRAIGHT)
		tree.back() = BRANCH;
}

void DumpTree::annotated_branch(const string &annotation, Node &node)
{
	append(annotation);
	begin_sub();
	last_branch();
	node.visit(*this);
	end_sub();
}

unsigned DumpTree::get_label(const Node &node)
{
	unsigned &label = node_labels[&node];
	if(!label)
		label = node_labels.size();
	return label;
}

string DumpTree::format_type(TypeDeclaration *type)
{
	return (type ? type->name : "?");
}

template<typename T>
typename T::const_iterator DumpTree::increment(typename T::const_iterator &iter, const T &container)
{
	typename T::const_iterator ret = iter++;
	if(iter==container.end())
		last_branch();
	return ret;
}

void DumpTree::visit(Block &block)
{
	append(format("Block %s", (block.use_braces ? "{}" : "(inline)")));
	begin_sub();

	for(std::map<string, VariableDeclaration *>::const_iterator i=block.variables.begin(); i!=block.variables.end(); ++i)
		append(format("Variable: %%%d %s %s", get_label(*i->second), i->second->type, i->first));

	bool labeled_body = !block.variables.empty();
	if(labeled_body)
	{
		last_branch();
		append("Body");
		begin_sub();
	}
	for(NodeList<Statement>::const_iterator i=block.body.begin(); i!=block.body.end(); )
	{
		NodeList<Statement>::const_iterator j = increment(i, block.body);
		(*j)->visit(*this);
	}
	if(labeled_body)
		end_sub();

	end_sub();
}

void DumpTree::visit(Literal &literal)
{
	append(format("Literal: %s -> %s", literal.token, format_type(literal.type)));
}

void DumpTree::visit(ParenthesizedExpression &parexpr)
{
	annotated_branch(format("(expr) -> %s", format_type(parexpr.type)), *parexpr.expression);
}

void DumpTree::visit(VariableReference &var)
{
	string text;
	if(var.declaration)
		text += format("%%%d ", get_label(*var.declaration));
	text += format("%s (var) -> %s", var.name, format_type(var.type));
	append(text);
}

void DumpTree::visit(InterfaceBlockReference &iface)
{
	string text;
	if(iface.declaration)
		text += format("%%%d ", get_label(*iface.declaration));
	text += format("%s (iface)", iface.name);
	append(text);
}

void DumpTree::visit(MemberAccess &memacc)
{
	string text = "Member access:";
	if(memacc.declaration)
		text += format(" %%%d", get_label(*memacc.declaration));
	text += format(" .%s -> %s", memacc.member, format_type(memacc.type));
	annotated_branch(text, *memacc.left);
}

void DumpTree::visit(UnaryExpression &unary)
{
	string text = format("Unary: %s, %sfix -> %s", unary.oper->token, (unary.oper->type==Operator::PREFIX ? "pre" : "post"), format_type(unary.type));
	annotated_branch(text, *unary.expression);
}

void DumpTree::visit(BinaryExpression &binary)
{
	append(format("Binary: %s -> %s", (binary.oper->token[0]=='[' ? "[]" : binary.oper->token), format_type(binary.type)));
	begin_sub();
	binary.left->visit(*this);
	last_branch();
	binary.right->visit(*this);
	end_sub();
}

void DumpTree::visit(Assignment &assign)
{
	append(format("Assignment: %s%s", assign.oper->token, (assign.self_referencing ? " (self-referencing)" : "")));
	begin_sub();
	if(assign.target_declaration)
		append(format("Target: %%%d %s %s", get_label(*assign.target_declaration), assign.target_declaration->type, assign.target_declaration->name));
	assign.left->visit(*this);
	last_branch();
	assign.right->visit(*this);
	end_sub();
}

void DumpTree::visit(FunctionCall &call)
{
	string head = "Function call: ";
	if(call.declaration)
		head += format("%%%d ", get_label(*call.declaration));
	head += call.name;
	if(call.constructor)
		head += " (constructor)";
	head += format(" -> %s", format_type(call.type));
	append(head);

	begin_sub();
	for(NodeArray<Expression>::const_iterator i=call.arguments.begin(); i!=call.arguments.end(); )
	{
		NodeArray<Expression>::const_iterator j = increment(i, call.arguments);
		(*j)->visit(*this);
	}
	end_sub();
}

void DumpTree::visit(ExpressionStatement &expr)
{
	annotated_branch("expr;", *expr.expression);
}

void DumpTree::visit(Import &import)
{
	append(format("import %s", import.module));
}

void DumpTree::visit(Precision &prec)
{
	append(format("precision %s %s", prec.precision, prec.type));
}

void DumpTree::visit(Layout &layout)
{
	append("Layout");
	begin_sub();
	for(vector<Layout::Qualifier>::const_iterator i=layout.qualifiers.begin(); i!=layout.qualifiers.end(); )
	{
		vector<Layout::Qualifier>::const_iterator j = increment(i, layout.qualifiers);
		string qualifier = j->name;
		if(j->has_value)
			qualifier += format("=%d", j->value);
		append(qualifier);
	}
	end_sub();
}

void DumpTree::visit(InterfaceLayout &layout)
{
	annotated_branch(format("Layout: %s", layout.interface), layout.layout);
}

void DumpTree::visit(BasicTypeDeclaration &type)
{
	append(format("%%%d typedef %s", get_label(type), type.name));
	begin_sub();
	if(type.kind!=BasicTypeDeclaration::VECTOR && type.kind!=BasicTypeDeclaration::MATRIX)
		last_branch();
	if(type.base_type)
		append(format("%s: %%%d %s", (type.kind==BasicTypeDeclaration::ALIAS ? "Alias of" : "Base"), get_label(*type.base_type), type.base_type->name));

	last_branch();
	if(type.kind==BasicTypeDeclaration::VECTOR)
		append(format("Vector: %d", type.size));
	else if(type.kind==BasicTypeDeclaration::MATRIX)
		append(format("Matrix: %dx%d", type.size&0xFFFF, type.size>>16));
	end_sub();
}

void DumpTree::visit(ImageTypeDeclaration &type)
{
	append(format("%%%d typedef %s", get_label(type), type.name));
	begin_sub();

	if(!type.shadow && !type.base_type)
		last_branch();
	static const char *dims[] = { "1D", "2D", "3D", "Cube" };
	append(format("Dimensions: %s%s", dims[type.dimensions-1], (type.array ? " array" : "")));

	if(!type.shadow)
		last_branch();
	if(type.base_type)
		append(format("Element type: %%%d %s", get_label(*type.base_type), type.base_type->name));

	last_branch();
	if(type.shadow)
		append("Shadow");

	end_sub();
}

void DumpTree::visit(StructDeclaration &strct)
{
	annotated_branch(format("%%%d struct %s", get_label(strct), strct.name), strct.members);
}

void DumpTree::visit(VariableDeclaration &var)
{
	string decl = format("%%%d ", get_label(var));
	if(var.constant)
		decl += "const ";
	if(!var.interpolation.empty())
		decl += format("%s ", var.interpolation);
	if(!var.sampling.empty())
		decl += format("%s ", var.sampling);
	if(!var.interface.empty())
		decl += format("%s ", var.interface);
	if(!var.precision.empty())
		decl += format("%s ", var.precision);
	decl += format("%s %s", var.type, var.name);
	if(var.source==BUILTIN_SOURCE)
		decl += " (builtin)";
	else if(var.linked_declaration)
		decl += " (linked)";
	append(decl);

	begin_sub();
	if(!var.layout && !var.array && !var.init_expression)
		last_branch();
	if(var.type_declaration)
		append(format("Type: %%%d %s", get_label(*var.type_declaration), var.type_declaration->name));

	if(!var.array && !var.init_expression)
		last_branch();
	if(var.layout)
		var.layout->visit(*this);

	if(!var.init_expression)
		last_branch();
	if(var.array)
	{
		if(var.array_size)
			annotated_branch("Array []", *var.array_size);
		else
			append("Array []");
	}

	last_branch();
	if(var.init_expression)
		var.init_expression->visit(*this);
	end_sub();
}

void DumpTree::visit(InterfaceBlock &block)
{
	string head;
	if(!block.instance_name.empty())
		head += format("%%%d ", get_label(block));
	head += format("%s %s", block.interface, block.name);
	if(!block.instance_name.empty())
		head += format(" %s", block.instance_name);
	if(block.array)
		head += "[]";
	if(block.source==BUILTIN_SOURCE)
		head += " (builtin)";
	else if(block.linked_block)
		head += " (linked)";
	annotated_branch(head, block.members);
}

void DumpTree::visit(FunctionDeclaration &func)
{
	string text = format("%%%d %s %s", get_label(func), func.return_type, func.name);
	if(func.source==BUILTIN_SOURCE)
		text += " (builtin)";
	else if(!func.definition)
		text += " (undefined)";
	append(text);
	begin_sub();
	for(NodeArray<VariableDeclaration>::const_iterator i=func.parameters.begin(); i!=func.parameters.end(); ++i)
		(*i)->visit(*this);
	last_branch();
	if(func.definition==&func)
		func.body.visit(*this);
	else if(func.definition)
		append(format("Definition: %%%d", get_label(*func.definition)));
	end_sub();
}

void DumpTree::visit(Conditional &cond)
{
	append("if()");
	begin_sub();
	cond.condition->visit(*this);
	if(cond.else_body.body.empty())
		last_branch();
	cond.body.visit(*this);
	if(!cond.else_body.body.empty())
	{
		last_branch();
		cond.else_body.visit(*this);
	}
	end_sub();
}

void DumpTree::visit(Iteration &iter)
{
	append("for()");
	begin_sub();

	if(iter.init_statement)
		annotated_branch("Initialization", *iter.init_statement);
	if(iter.condition)
		annotated_branch("Condition", *iter.condition);
	if(iter.loop_expression)
		annotated_branch("Loop", *iter.loop_expression);
	last_branch();
	annotated_branch("Body", iter.body);

	end_sub();
}

void DumpTree::visit(Passthrough &pass)
{
	append("passthrough");
	if(pass.subscript)
	{
		begin_sub();
		last_branch();
		pass.subscript->visit(*this);
		end_sub();
	}
}

void DumpTree::visit(Return &ret)
{
	if(ret.expression)
		annotated_branch("return", *ret.expression);
	else
		append("return;");
}

void DumpTree::visit(Jump &jump)
{
	append(format("%s;", jump.keyword));
}

} // namespace SL
} // namespace GL
} // namespace Msp
