#ifndef MSP_GL_SL_GENERATE_H_
#define MSP_GL_SL_GENERATE_H_

#include <map>
#include <set>
#include <string>
#include <vector>
#include "visitor.h"

namespace Msp {
namespace GL {
namespace SL {

/** Combines multiple declarations of the same identifier into one. */
class DeclarationCombiner: private TraversingVisitor
{
private:
	std::map<std::string, std::vector<FunctionDeclaration *> > functions;
	std::map<std::string, VariableDeclaration *> variables;
	std::set<Node *> nodes_to_remove;

public:
	void apply(Stage &);

private:
	virtual void visit(Block &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(FunctionDeclaration &) { }
};

/** Manipulates specialization constants.  If values are specified, turns
specialization constants into normal constants.  Without values assigns
automatic constant_ids to specialization constants. */
class ConstantSpecializer: private TraversingVisitor
{
private:
	const std::map<std::string, int> *values;

public:
	ConstantSpecializer();

	void apply(Stage &, const std::map<std::string, int> *);

private:
	virtual void visit(VariableDeclaration &);
};

/** Forms links between nested blocks in the syntax tree. */
class BlockHierarchyResolver: private TraversingVisitor
{
public:
	void apply(Stage &s) { s.content.visit(*this); }

private:
	virtual void enter(Block &);
};

/** Resolves types of variables and base types of other types. */
class TypeResolver: private TraversingVisitor
{
private:
	Stage *stage;
	std::map<TypeDeclaration *, TypeDeclaration *> alias_map;
	std::map<TypeDeclaration *, TypeDeclaration *> array_types;
	NodeList<Statement>::iterator type_insert_point;

public:
	TypeResolver();

	void apply(Stage &);

private:
	TypeDeclaration *resolve_type(const std::string &);
	virtual void visit(Block &);
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(FunctionDeclaration &);
};

/** Resolves variable references.  Variable references which match the name
of an interface block are turned into interface block references. */
class VariableResolver: private TraversingVisitor
{
private:
	Stage *stage;
	std::map<std::string, VariableDeclaration *> *r_members;
	RefPtr<InterfaceBlockReference> r_iface_ref;
	std::string block_interface;
	bool record_target;
	bool r_self_referencing;
	VariableDeclaration *r_assignment_target;

public:
	VariableResolver();

	void apply(Stage &);

private:
	virtual void enter(Block &);
	virtual void visit(VariableReference &);
	virtual void visit(InterfaceBlockReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(FunctionCall &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
};

/** Resolves types and lvalueness of expressions. */
class ExpressionResolver: private TraversingVisitor
{
private:
	enum Compatibility
	{
		NOT_COMPATIBLE,
		LEFT_CONVERTIBLE,
		RIGHT_CONVERTIBLE,
		SAME_TYPE
	};

	Stage *stage;
	std::vector<BasicTypeDeclaration *> basic_types;

public:
	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	static bool is_scalar(BasicTypeDeclaration &);
	static bool is_vector_or_matrix(BasicTypeDeclaration &);
	static BasicTypeDeclaration *get_element_type(BasicTypeDeclaration &);
	static bool can_convert(BasicTypeDeclaration &, BasicTypeDeclaration &);
	static Compatibility get_compatibility(BasicTypeDeclaration &, BasicTypeDeclaration &);
	BasicTypeDeclaration *find_type(BasicTypeDeclaration::Kind, unsigned);
	BasicTypeDeclaration *find_type(BasicTypeDeclaration &, BasicTypeDeclaration::Kind, unsigned);
	void convert_to(RefPtr<Expression> &, BasicTypeDeclaration &);
	bool convert_to_element(RefPtr<Expression> &, BasicTypeDeclaration &);

	virtual void visit(Literal &);
	virtual void visit(ParenthesizedExpression &);
	virtual void visit(VariableReference &);
	virtual void visit(InterfaceBlockReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(FunctionCall &);
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(VariableDeclaration &);
};

/** Resolves function declarations and calls. */
class FunctionResolver: private TraversingVisitor
{
private:
	Stage *stage;
	std::map<std::string, std::vector<FunctionDeclaration *> > declarations;

public:
	void apply(Stage &);

private:
	virtual void visit(FunctionCall &);
	virtual void visit(FunctionDeclaration &);
};

/** Materializes implicitly declared interfaces.

Out variable declarations inside functions are moved to the global scope.

Passthrough statements are processed, generating out variables to match in
variables and copying values.

Unresolved variables are looked up in the previous stage's out variables. */
class InterfaceGenerator: private TraversingVisitor
{
private:
	Stage *stage;
	std::string in_prefix;
	std::string out_prefix;
	bool function_scope;
	InterfaceBlock *iface_block;
	bool copy_block;
	Block *iface_target_block;
	NodeList<Statement>::iterator iface_insert_point;
	NodeList<Statement>::iterator assignment_insert_point;
	std::set<Node *> nodes_to_remove;

public:
	InterfaceGenerator();

	void apply(Stage &);

private:
	static std::string get_out_prefix(Stage::Type);
	std::string change_prefix(const std::string &, const std::string &) const;
	virtual void visit(Block &);
	VariableDeclaration *generate_interface(VariableDeclaration &, const std::string &, const std::string &);
	InterfaceBlock *generate_interface(InterfaceBlock &);
	ExpressionStatement &insert_assignment(const std::string &, Expression *);
	virtual void visit(VariableReference &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Passthrough &);
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
