#include <msp/datafile/builtinsource.h>
#include <msp/fs/utils.h>
#include "animation.h"
#include "armature.h"
#include "camera.h"
#include "font.h"
#include "keyframe.h"
#include "lighting.h"
#include "material.h"
#include "mesh.h"
#include "module.h"
#include "object.h"
#include "pipelinetemplate.h"
#include "pose.h"
#include "program.h"
#include "resourcemanager.h"
#include "resources.h"
#include "sampler.h"
#include "technique.h"
#include "texture1d.h"
#include "texture2d.h"
#include "texture2darray.h"
#include "texturecube.h"
#include "glsl/compiler.h"

using namespace std;

namespace Msp {
namespace GL {

void init_shaderlib(DataFile::BuiltinSource &);
void init_builtin_data(DataFile::BuiltinSource &);

Resources::Resources():
	default_tex_filter(Texture::can_generate_mipmap() ? LINEAR_MIPMAP_LINEAR : LINEAR),
	default_tex_anisotropy(1.0f),
	srgb_conversion(false),
	resource_manager(0)
{
	add_type<Animation>().suffix(".anim").keyword("animation");
	add_type<Armature>().suffix(".arma").keyword("armature");
	add_type<Camera>().keyword("camera");
	add_type<Font>().keyword("font");
	add_type<KeyFrame>().suffix(".kframe").keyword("keyframe");
	add_type<Lighting>().suffix(".lightn").keyword("lighting");
	add_type<Material>().suffix(".mat").creator(&Resources::create_material);
	add_type<Mesh>().keyword("mesh").creator(&Resources::create_mesh);
	add_type<Module>().suffix(".glsl").creator(&Resources::create_module);
	add_type<Object>().keyword("object");
	add_type<PipelineTemplate>().suffix(".pipe").keyword("pipeline");
	add_type<Pose>().keyword("pose");
	add_type<Program>().keyword("shader").creator(&Resources::create_program);
	add_type<Sampler>().suffix(".samp").keyword("sampler");
	add_type<Technique>().suffix(".tech").keyword("technique");
	add_type<Texture1D>().base<Texture>().suffix(".tex1d").keyword("texture1d");
	add_type<Texture2D>().base<Texture>().suffix(".tex2d").suffix(".png").suffix(".jpg").keyword("texture2d").creator(&Resources::create_texture2d);
	add_type<Texture3D>().base<Texture>().suffix(".tex3d").keyword("texture3d");
	add_type<TextureCube>().base<Texture>().suffix(".texcb").keyword("texture_cube");
	add_type<Texture2DArray>().base<Texture>().suffix(".tex2da").keyword("texture2d_array");

	add_source(get_builtins());
}

const DataFile::CollectionSource &Resources::get_builtins()
{
	static DataFile::BuiltinSource builtins;
	bool init_done = false;

	if(!init_done)
	{
		init_builtin_data(builtins);
		init_shaderlib(builtins);
		init_done = true;
	}

	return builtins;
}

void Resources::set_default_texture_filter(TextureFilter tf)
{
	default_tex_filter = tf;
}

void Resources::set_default_texture_anisotropy(float a)
{
	default_tex_anisotropy = a;
}

void Resources::set_srgb_conversion(bool c)
{
	srgb_conversion = c;
}

void Resources::set_resource_manager(ResourceManager *m)
{
	resource_manager = m;
}

Material *Resources::create_material(const string &name)
{
	if(RefPtr<IO::Seekable> io = open_raw(name))
	{
		DataFile::Parser parser(*io, name);
		Material::GenericLoader ldr(this);
		ldr.load(parser);
		return ldr.get_material();
	}

	return 0;
}

Mesh *Resources::create_mesh(const string &name)
{
	if(!resource_manager)
		return 0;

	if(RefPtr<IO::Seekable> io = open_raw(name))
	{
		RefPtr<Mesh> mesh = new Mesh(resource_manager);
		resource_manager->set_resource_location(*mesh, *this, name);
		return mesh.release();
	}

	return 0;
}

Texture2D *Resources::create_texture2d(const string &name)
{
	string ext = FS::extpart(name);
	if(ext==".tex2d" && !resource_manager)
		return 0;

	if(RefPtr<IO::Seekable> io = open_raw(name))
	{
		Graphics::Image image;
		if(!resource_manager)
			image.load_io(*io);

		RefPtr<Texture2D> tex = new Texture2D(resource_manager);

		if(ext==".tex2d")
		{
			DataFile::Parser parser(*io, name);
			Texture2D::Loader ldr(*tex, *this);
			ldr.load(parser);
		}
		else
		{
			Sampler &samp = tex->get_default_sampler();
			if(is_mipmapped(default_tex_filter))
			{
				tex->set_auto_generate_mipmap(true);
				samp.set_mag_filter(LINEAR);
			}
			else
				samp.set_mag_filter(default_tex_filter);
			samp.set_min_filter(default_tex_filter);
			samp.set_max_anisotropy(default_tex_anisotropy);
		}

		if(resource_manager)
			resource_manager->set_resource_location(*tex, *this, name);
		else
			tex->image(image);
		return tex.release();
	}

	return 0;
}

Module *Resources::create_module(const string &name)
{
	if(RefPtr<IO::Seekable> io = open_raw(name))
	{
		RefPtr<Module> module = new Module;
		module->load_source(*io, this, name);
		return module.release();
	}

	return 0;
}

Program *Resources::create_program(const string &name)
{
	string ext = FS::extpart(name);
	string base = FS::basepart(name);
	string ext2 = FS::extpart(base);
	if(ext==".shader" && ext2==".glsl")
	{
		Module &module = get<Module>(base);
		RefPtr<Program> shprog = new Program;
		shprog->add_stages(module);
		shprog->link();
		return shprog.release();
	}

	return 0;
}

} // namespace GL
} // namespace Msp
